
/*
 * xprompt - interactively prompt for input
 *
 * This program is mainly intended to be used by scripts to prompt
 * the user for text input and return the response(s), printed to stdout
 * one per line.
 *
 * Basic editing functionality is supported, as is the ability to mask
 * the input text so that sensitive material, such as a password, isn't
 * easily read by a third party.
 * Other than suppressing a core dump, little else is done to guard against
 * potentially sensitive prompt or reply strings being leaked and
 * no guarantees are provided with respect to the secure handling of
 * sensitive information.  Use accordingly.
 *
 * 13-Aug-2019 Convert to X11R7, add support for a variety of echo modes
 *             (e.g., for entry of passwords), and allow reply text to be pasted
 *             Remove the "wordchars" functionality for erasing words.
 *             This major revision was implemented by the original author,
 *             who is no longer affiliated with U.B.C.
 *             This major revision (i.e., xprompt 2.x and later) uses the
 *             2-Clause BSD License, but the original, slightly less permissive
 *             MIT-type licensing terms should be honoured.
 *             Both licenses appear below and in the LICENSE file.
 *             Barry Brachman <brachman@dss.ca>, Distributed Systems Software
 *
 * 28-Sep-1991 Convert to X11R5, retain X11R4, lose X11R3
 *             Casey Leedom <casey@gauss.llnl.gov>
 *
 * 28-Jan-1990 Convert to X11R4, retain X11R3 compatibility
 *
 * 11-Mar-1989 Fixed problem with XSetInputFocus being done before windows
 *             were viewable.
 *
 * 24-Jan-1989 Written for X11R3.  Released for distribution.
 *
 *
 * Copyright (c) 2019 Distributed Systems Software.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Copyright (c) 1989, 1990, 1991, 1995 by
 * Barry Brachman and the University of British Columbia
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of U.B.C. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  U.B.C. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * -----
 * [Obsolete contact info for the original version/author:]
 * Barry Brachman <brachman@cs.ubc.ca>
 * Dept. of Computer Science
 * 2366 Main Mall, Univ. of British Columbia
 * Vancouver, B.C. V6T 1Z4
 */

#ifndef lint
static const char revid[] =
  "$Id: xprompt.c 3098 2019-10-18 21:34:17Z brachman $";
#endif

#include <sys/types.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xatom.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/XawInit.h>

#include "version.h"

#ifndef MAYBE_UNUSED
#define MAYBE_UNUSED    __attribute__ ((__unused__))
#endif

#define MAXIMUM(x, y)		((x) > (y) ? (x) : (y))

enum {
  GEOM_WIDTH_PAD				= 25,		/* Extra padding for window width. */
  GEOM_HEIGHT_PAD				= 15,		/* Same for height. */
  MAX_PROMPTS                   = 99,		/* Arbitrary, but sufficient. */
  DEFAULT_GEOM_X                = 500,
  DEFAULT_GEOM_Y                = 400,
  DEFAULT_INSIDE_BORDER_WIDTH	= 1,
  DEFAULT_MAXIMUM_REPLY_LEN		= 40,	/* Excludes the terminating null byte. */
  DEFAULT_RETURN_EXIT           = TRUE,
  DEFAULT_WARP                  = FALSE,
  DEFAULT_GRAB                  = TRUE,
  DEFAULT_DEBUG                 = FALSE,
  DEFAULT_DMASK_DURATION		= 1000	/* Delayed mask time, in ms. */
};

/*
 * There are several input echoing modes, mainly to support input of sensitive
 * information, such as passwords.  The echo mode can be selected by a
 * command line flag.
 *
 * Echo modes are:
 * 1) ECHO_ON: Normal echo (the default)
 * 2) ECHO_MASKED:  Masked echo; each echoed input character is presented as
 *    echo_mask_char, a '*' or some other character configured at compile time
 * 3) ECHO_DMASKED: Delayed masked echo; each input character is echoed as
 *    itself, then replaced by the masking character (echo_dmask_char) either
 *    a) immediately after the next input character is read or b) after a
 *    configurable delay (DEFAULT_DMASK_DURATION), whichever happens first
 * 4) ECHO_OFF: No echo
 *
 * There is an echo mode toggle function.
 * Also, regardless of the echo mode, there are copy-to-clipboard and
 * paste-from-clipboard functions.
 */

typedef enum Echo_mode {
  ECHO_OFF          = 0,		/* No echoing. */
  ECHO_ON           = 1,		/* Normal echo. */
  ECHO_MASKED       = 2,		/* Mask all input by echoing using ECHO_CHAR. */
  ECHO_DMASKED      = 3,		/* Echo normally, then mask using ECHO_CHAR. */
  ECHO_INVALID      = -1,		/* An invalid mode. */
  DEFAULT_ECHO_MODE = ECHO_ON
} Echo_mode;

#define DEFAULT_ECHO_MASK_CHAR		'*'
#define DEFAULT_ECHO_DMASK_CHAR		'*'
#define DEFAULT_ECHO_MODE_STR	"on"

#define MULTI_PROMPT_FMT	"%s[%d/%d]:"
#define SINGLE_PROMPT_FMT	"%s:"

static inline MAYBE_UNUSED int
streq(const char *p, const char *q)
{

    return(strcmp(p, q) == 0);
}

static inline MAYBE_UNUSED int
strcaseeq(const char *p, const char *q)
{

    return(strcasecmp(p, q) == 0);
}

/*
 * Command line options.
 * The fourth (last) field of each XrmoptionNoArg entry is the value to use if
 * the command line flag is present.
 */
static XrmOptionDescRec option_table[] = {
    {"-debug",   "Debug",               XrmoptionNoArg,  (XPointer) "on" },
	{"-dmd",     "dmaskDuration",       XrmoptionSepArg, NULL },
    {"-echo",    "Echo",                XrmoptionSepArg, NULL },
	{"-efn",     "echoFont",            XrmoptionSepArg, NULL },
	{"-grab",    "Grab",                XrmoptionNoArg,  (XPointer) "on" },
	{"-ibw",     "insideborderWidth",   XrmoptionSepArg, NULL },
	{"-nodebug", "Debug",               XrmoptionNoArg,  (XPointer) "off" },
	{"-nograb",  "Grab",                XrmoptionNoArg,  (XPointer) "off" },
	{"-nowarp",  "Warp",                XrmoptionNoArg,  (XPointer) "off" },
	{"-p",       "",                    XrmoptionSkipLine, NULL },
	{"-pfn",     "promptFont",          XrmoptionSepArg, NULL },
	{"-re",      "returnExit",          XrmoptionNoArg,  (XPointer) "on" },
	{"-nore",    "returnExit",          XrmoptionNoArg,  (XPointer) "off" },
	{"-rlen",    "Rlen",                XrmoptionSepArg, NULL },
	{"-tf",      "textTranslationFile", XrmoptionSepArg, NULL },
	{"-warp",    "Warp",                XrmoptionNoArg, (XPointer) "on" },
};

typedef struct {
  Boolean debug;			/* If TRUE, emit debugging messages to stderr. */
  char *echo_mode_str;		/* Echo mode flag, as a string. */
  char *echo_mask_char;		/* Mask character for ECHO_MASKED, as a string. */
  char *echo_dmask_char;	/* Mask character for ECHO_DMASKED, as a string. */
  int rlen;					/* Maximum reply length. */
  Boolean grab;				/* If TRUE, grab the keyboard. */
  Boolean warp;				/* If TRUE, warp to reply window. */
  Boolean return_exit;		/* If TRUE, exit on <return> if single prompt. */
  int dmask_duration;		/* Delay in ms for dmask echo mode. */
  char *texttranslations;
  char *texttranslationfile;
  char *geometry;
  int borderwidth;
  int insideborderwidth;
  XFontStruct *font;
  XFontStruct *pfont;
  XFontStruct *efont;
} app_resourceRec, *app_res;

static app_resourceRec app_resources;

static XtResource resources[] = {
  {"debug",  "Debug", XtRBoolean, sizeof(Boolean),
   XtOffset(app_res, debug), XtRImmediate, (XPointer) DEFAULT_DEBUG },
  {"rlen",  "Rlen", XtRInt, sizeof(int),
   XtOffset(app_res, rlen), XtRImmediate,
   (XPointer) DEFAULT_MAXIMUM_REPLY_LEN},
  {"grab",  "Grab", XtRBoolean, sizeof(Boolean),
   XtOffset(app_res, grab), XtRImmediate, (XPointer) DEFAULT_GRAB },

  {"echo",   "Echo", XtRString, sizeof(char *),
   XtOffset(app_res, echo_mode_str), XtRString,
   (XPointer) DEFAULT_ECHO_MODE_STR },
  {"dmd",    "dmaskDuration", XtRInt, sizeof(int),
   XtOffset(app_res, dmask_duration), XtRImmediate,
   (XPointer) DEFAULT_DMASK_DURATION },
  {"insideborderwidth", "insideborderWidth", XtRInt, sizeof(int),
   XtOffset(app_res, insideborderwidth), XtRImmediate,
   (XPointer) DEFAULT_INSIDE_BORDER_WIDTH},

  {"echomaskchar",  "echoMaskChar", XtRString, sizeof(XPointer),
   XtOffset(app_res, echo_mask_char), XtRString,
   (XPointer) NULL },
  {"echodmaskchar",  "echoDmaskChar", XtRString, sizeof(XPointer),
   XtOffset(app_res, echo_dmask_char), XtRString,
   (XPointer) NULL },

  {"echofont", "echoFont", XtRFontStruct, sizeof(XFontStruct *),
   XtOffset(app_res, efont), XtRString, NULL},
  {"promptfont", "promptFont", XtRFontStruct, sizeof(XFontStruct *),
   XtOffset(app_res, pfont), XtRString, NULL},
  {"returnexit", "returnExit", XtRBoolean, sizeof(Boolean),
   XtOffset(app_res, return_exit), XtRImmediate,
   (XPointer) DEFAULT_RETURN_EXIT},
  {"texttranslationfile", "textTranslationFile", XtRString, sizeof(XPointer),
   XtOffset(app_res, texttranslationfile), XtRString, NULL},
  {"texttranslations", "textTranslations", XtRString, sizeof(XPointer),
   XtOffset(app_res, texttranslations), XtRString, NULL},
  {"warp",  "Warp", XtRBoolean, sizeof(Boolean),
   XtOffset(app_res, warp), XtRImmediate, (XPointer) DEFAULT_WARP },
  {XtNgeometry,  "Geometry", XtRString, sizeof(XPointer),
   XtOffset(app_res, geometry), XtRString, (XPointer) "+500+400"},
  {XtNborderWidth, "borderWidth", XtRInt, sizeof(int),
   XtOffset(app_res, borderwidth), XtRImmediate, (XPointer) 1},
  {XtNfont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
   XtOffset(app_res, font), XtRString, (XPointer) XtDefaultFont},
};

/*
 * Map events to actions.
 * We need to have our own bindings because of our implementation of
 * "raw" echoing modes.  Only a small subset of the normal actions should
 * be needed for this application, because it operates on just one (typically
 * short) line of text at a time, but it should be easy to add new actions
 * as required.
 *
 * Custom event handler names are prefixed with "xp-" to avoid confusion
 * with similar X Toolkit actions.
 *
 * See: Athena Widget Set - C Language Interface, Chapter 5 Text Widgets,
 * Default Translation Bindings.
 */
static char xprompt_TextTranslations[] =
"\
Ctrl<Key>A:         xp-beginning-of-line() \n\
Ctrl<Key>C:         xp-abort() \n\
Ctrl<Key>D:         xp-finish-prompt() \n\
Ctrl<Key>E:         xp-end-of-line() \n\
Ctrl<Key>H:         xp-erase-char() \n\
Ctrl<Key>J:         xp-next-prompt() \n\
Ctrl<Key>K:         xp-kill-to-end-of-line() \n\
Ctrl<Key>L:         xp-redraw-display() \n\
Ctrl<Key>M:         xp-next-prompt() \n\
Ctrl<Key>T:         xp-toggle-echo-mode() \n\
Ctrl<Key>U:         xp-erase-line() \n\
Ctrl<Key>W:         xp-erase-word() \n\
<Key>Down:          xp-next-prompt() \n\
<Key>Up:            xp-previous-prompt() \n\
<Key>Linefeed:      xp-next-prompt() \n\
<Key>Return:        xp-next-prompt-or-finish() \n\
\
<Btn1Up>:           xp-select-reply() \n\
<Btn1Down>:         xp-ignore() \n\
<Btn1Motion>:       xp-ignore() \n\
\
<Btn2Up>:           xp-ignore() \n\
<Btn2Down>:         xp-paste-selection() \n\
<Btn2Motion>:       xp-ignore() \n\
\
<Btn3Up>:           xp-ignore() \n\
<Btn3Down>:         xp-ignore() \n\
<Btn3Motion>:       xp-ignore() \n\
\
<VisibilityNotify>: visibility-event() \n\
<Key>Left:          xp-backward-char() \n\
<Key>Right:         xp-forward-char() \n\
<Key>BackSpace:	    xp-erase-char() \n\
<Key>Delete:        xp-delete-char() \n\
<Key>:              xp-insert-char() \n\
";

typedef struct prompt_args Prompt_args;
struct prompt_args {
  char *prompt_buf;		/* Text string to query (label) this prompt. */
  char *reply_buf;		/* The user's response to the prompt. */
  Boolean seenprompt;	/* TRUE if this prompt has been presented. */
  Prompt_args *next;	/* If non-NULL, the next prompt in the list. */
  Prompt_args *prev;	/* If non-NULL, the previous prompt in the list. */
};
static Prompt_args *prompt_args, *npa, *npa_prev;

/* The current prompt's number (from 1 to nprompts). */
static int current_prompt;
/* The total number of prompts to be shown. */
static int nprompts;
/* The number of prompts shown to the user so far. */
static int nprompts_seen;

/* The current echo, prompt and reply strings. */
static char *echo_buf, *prompt_buf, *reply_buf;

static Widget toplevel_w, echo_w, prompt_w;

static XtIntervalId echo_dmask_timer;
static XtAppContext app_context;

/* The original cursor position, if a warp is requested. */
struct warp_info {
  Window orig_win;
  int orig_x, orig_y;
} warp_info;

static long echo_dmask_pos = -1;
/*
 * One might want different masking characters to indicate which
 * echo mode is in effect.
 */
static int echo_mask_char = DEFAULT_ECHO_MASK_CHAR;
static int echo_dmask_char = DEFAULT_ECHO_DMASK_CHAR;

/*
 * The insertion point is the index within reply_buf of where the next
 * input character should be placed.
 * When the insertion point is at the end of reply_buf there should
 * always be a null byte there.
 * An XawTextPosition is a "long int", so a value of -1 is a possibility.
 */
static long insertion_point = 0;

/* Non-zero if the prompt window is visible to the user. */
static int visible;

/* The maximum length of a prompt string, computed at run-time. */
static int max_displayed_prompt_len = 0;

static int debug_flag = 0;
static Echo_mode echo_mode = DEFAULT_ECHO_MODE;
static char *progname = NULL;

/*
 * The semantics of setrlimit(2) do not allow limits to be
 * increased, so once the core dump is disabled we cannot change our mind.
 */
void
disable_dump(void)
{
  struct rlimit rlimit;

  rlimit.rlim_cur = rlimit.rlim_max = 0;

  /* This is a POSIX.1-2001 system call. */
  if (setrlimit(RLIMIT_CORE, &rlimit) == -1) {
    fprintf(stderr, "%s: setrlimit() failed", progname);
    exit(1);
  }
}

/*
 * Convert EM to its canonical string equivalent, for printing.
 */
static char *
echo_mode_str(Echo_mode em)
{
  char *str;

  switch (em) {
  case ECHO_ON:
	str = "on";
	break;
  case ECHO_OFF:
	str = "off";
	break;
  case ECHO_MASKED:
	str = "mask";
	break;
  case ECHO_DMASKED:
	str = "dmask";
	break;
  case ECHO_INVALID:
  default:
	str = "???";
	break;
  }

  return(str);
}

/*
 * Validate ARG as an echo mode.
 */
static Echo_mode
echo_mode_arg(char *arg)
{
  Echo_mode mode;

  if (strcaseeq(arg, "on"))
	mode = ECHO_ON;
  else if (strcaseeq(arg, "off"))
	mode = ECHO_OFF;
  else if (strcaseeq(arg, "mask"))
	mode = ECHO_MASKED;
  else if (strcaseeq(arg, "dmask"))
	mode = ECHO_DMASKED;
  else
	mode = ECHO_INVALID;

  return(mode);
}

/*
 * Copy string SRC to DST (of size DSTLEN bytes), after zeroing it.
 */
static void
set_buffer(void *dst, void *src, size_t dstlen)
{

  memset(dst, 0, dstlen);
  strcpy(dst, src);
}

static void
make_geometry_str(char *buf, size_t buflen, int mask, int w, int h, int x, int y)
{

  snprintf(buf, buflen, "%dx%d%c%d%c%d",
		   w, h,
		   (mask & XNegative) ? '-' : '+', x,
		   (mask & YNegative) ? '-' : '+', y);
}

/*
 * Tell widget W that the length or contents of its string resource BUF may
 * have changed.
 */
static void
refresh_widget_buffer(Widget w, char *buf)
{
  size_t len;
  Arg args[2];

  len = strlen(buf);
  if (debug_flag)
	fprintf(stderr, "refresh_widget_buffer: \"%s\", len=%ld\n", buf, len);

  XtSetArg(args[0], XtNstring, buf);
  XtSetArg(args[1], XtNlength, len);

  XtSetValues(w, args, (Cardinal) 2);
}

static int
displayed_prompt_length(int n, size_t plen)
{
  int len;
  char buf[1];

  buf[0] = '\0';
  if (n == 1)
	len = snprintf(buf, 0, SINGLE_PROMPT_FMT, "");
  else
	len = snprintf(buf, 0, MULTI_PROMPT_FMT, "", n, n);

  if (len == -1)
	return(-1);

  return(plen + len);
}

static int
set_prompt_str(char *dst_buf, char *src_buf, int which_prompt, int total_prompts)
{
  int len;

  if (total_prompts == 1)
	len = sprintf(dst_buf, SINGLE_PROMPT_FMT, src_buf);
  else
	len = sprintf(dst_buf, MULTI_PROMPT_FMT, src_buf, which_prompt,
				  total_prompts);
  
  return(len);
}

static void
set_echo_str(char *str)
{

  if (echo_mode == ECHO_ON)
	set_buffer(echo_buf, str, app_resources.rlen + 1);
  else if (echo_mode == ECHO_MASKED || echo_mode == ECHO_DMASKED) {
	size_t len;
	
	memset(echo_buf, 0, app_resources.rlen + 1);
	len = strlen(str);
	memset(echo_buf,
		   (echo_mode == ECHO_MASKED) ? echo_mask_char : echo_dmask_char, len);
  }
  else
	set_buffer(echo_buf, "", app_resources.rlen + 1);
}

static void
set_reply_str(char *reply_str)
{

  set_buffer(reply_buf, reply_str, app_resources.rlen + 1);
}

/*
 * A prompt consists of three components:
 * 1) The prompt string, which describes an input requested from the user
 * 2) The reply string, which is either a default given on the command line
 *    or input provided by the user
 * 3) The echo string, which is text displayed as the user enters the reply
 */
static void
set_prompt(char *prompt_str, char *reply_str, int which_prompt,
		   int total_prompts)
{

  set_prompt_str(prompt_buf, prompt_str, which_prompt, total_prompts);

  insertion_point = strlen(reply_str);

  set_echo_str(reply_str);
  set_reply_str(reply_str);
}

/*
 * Change the displayed prompt from NPA_PREV to NPA (prompt number
 * CURRENT_PROMPT).
 * Initially, and when we switch back to a previously displayed prompt, the
 * insertion position is set to the end of the reply buffer (i.e., it may not
 * be at the same place as when last the prompt was displayed).
 * Whether this is a bug or feature, I'm not sure it matters much.
 */
static void
exchange_prompt(void)
{

  if (debug_flag)
	fprintf(stderr, "exchange_prompt: to #%d\n", current_prompt);

  /*
   * Ignore any pending ECHO_DMASK timeout that might be scheduled for the
   * prompt we're switching away from.
   */
  echo_dmask_pos = -1;

  /* Save the reply to the current prompt. */
  strcpy(npa_prev->reply_buf, reply_buf);

  /*
   * Initialze (or reinitialize) the new current prompt and its echo string.
   */
  set_prompt(npa->prompt_buf, npa->reply_buf, current_prompt, nprompts);

  if (npa->seenprompt == FALSE) {
	npa->seenprompt = TRUE;
	nprompts_seen++;
  }

  refresh_widget_buffer(prompt_w, prompt_buf);
  if (debug_flag)
	fprintf(stderr, "Changed prompt string: \"%s\"\n", prompt_buf);

  refresh_widget_buffer(echo_w, echo_buf);
  if (debug_flag)
	fprintf(stderr, "Changed echo string: \"%s\"\n", echo_buf);

  if (echo_mode == ECHO_OFF)
	insertion_point = strlen(reply_buf);
  else
	XawTextSetInsertionPoint(echo_w, strlen(reply_buf));

  XawTextDisplay(echo_w);
  XawTextDisplay(prompt_w);
}

static void
unwarp(void)
{

  if (app_resources.warp && warp_info.orig_x != -1 && warp_info.orig_y != -1) {
	XWarpPointer(XtDisplay(echo_w), None, warp_info.orig_win,
				 0, 0, 0, 0, warp_info.orig_x, warp_info.orig_y);
	XFlush(XtDisplay(echo_w));
  }
}

static int
load_user_text_translations(char *filename)
{
  char *p;
  FILE *fp;
  struct stat statb;
  XtTranslations t;

  p = NULL;
  fp = NULL;

  if (debug_flag)
	fprintf(stderr, "Loading user text translations from \"%s\"\n", filename);

  if ((fp = fopen(filename, "r")) == NULL) {
	fprintf(stderr, "xprompt: fopen(): %s\n", strerror(errno));
	fprintf(stderr, "xprompt: Can't open translation file \"%s\".\n", filename); 
	return(-1);
  }

  if (fstat(fileno(fp), &statb) == -1) {
	fprintf(stderr, "xprompt: fstat(): %s\n", strerror(errno));
	fprintf(stderr, "xprompt: Can't stat translation file '%s'.\n", filename);
  err:
	if (p != NULL)
	  free(p);

	if (fp != NULL)
	  fclose(fp);

	return(-1);
  }

  if (statb.st_size == 0) {
	/* File is empty, no translations. */
	if (debug_flag)
	  fprintf(stderr, "User translation file is empty\n");
	fclose(fp);
	return(0);
  }

  if ((p = (char *) malloc((size_t) statb.st_size)) == NULL) {
	fprintf(stderr, "xprompt: mallloc(): %s\n", strerror(errno));
	fprintf(stderr, "xprompt: Can't malloc translation table.\n");
	goto err;
  }

  if (fread(p, statb.st_size, 1, fp) == 0) {
	fprintf(stderr, "xprompt: fread(): %s\n", strerror(errno));
	fprintf(stderr, "xprompt: error reading translation table.\n");
	goto err;
  }

  fclose(fp);
  fp = NULL;

  if (debug_flag) {
	fprintf(stderr, "User translations from \"%s\":\n", filename);
	fprintf(stderr, "%s\n", p);
  }

  if ((t = XtParseTranslationTable(p)) == NULL) {
	fprintf(stderr, "xprompt: error parsing translation table from \"%s\".\n",
			filename);
	goto err;
  }

  XtOverrideTranslations(echo_w, t);

  free(p);

  return(0);
}

/*ARGSUSED*/
static void
BackwardChar(Widget w, XEvent *event, String *args, Cardinal *nargs)
{
  size_t len;
  XawTextPosition start_pos;

  if (echo_mode == ECHO_OFF)
	start_pos = insertion_point;
  else
	start_pos = XawTextGetInsertionPoint(echo_w);

  len = strlen(echo_buf);
  if (debug_flag)
	fprintf(stderr, "BackwardChar @%ld/%ld\n", start_pos, len);

  if (start_pos > 0) {
	insertion_point = start_pos - 1;
	if (echo_mode != ECHO_OFF)
	  XawTextSetInsertionPoint(echo_w, start_pos - 1);
  }
}

/*ARGSUSED*/
static void
ForwardChar(Widget w, XEvent *event, String *args, Cardinal *nargs)
{
  size_t len;
  XawTextPosition start_pos;

  start_pos = XawTextGetInsertionPoint(echo_w);
  len = strlen(echo_buf);
  if (debug_flag)
	fprintf(stderr, "ForwardChar @%ld/%ld\n", start_pos, len);

  if (start_pos < len) {
	insertion_point = start_pos + 1;
	XawTextSetInsertionPoint(echo_w, start_pos + 1);
  }
}

/*
 *
 */
/*ARGSUSED*/
static void
RedrawDisplay(Widget w, XEvent *event, String *args, Cardinal *nargs)
{

  if (debug_flag)
	fprintf(stderr, "RedrawDisplay\n");

  refresh_widget_buffer(echo_w, echo_buf);

  if (echo_mode != ECHO_OFF)
	XawTextSetInsertionPoint(echo_w, insertion_point);

  XawTextDisplay(echo_w);
}

/*
 * Cycle through the echo modes, e.g., so that the user can temporarily
 * see a reply that is currently masked.
 */
/*ARGSUSED*/
static void
ToggleEchoMode(Widget w, XEvent *event, String *args, Cardinal *nargs)
{
  Echo_mode pem;

  if (debug_flag)
	fprintf(stderr, "ToggleEchoMode\n");

  pem = echo_mode;
  switch (echo_mode) {
  case ECHO_OFF:
	echo_mode = ECHO_ON;
	break;

  case ECHO_ON:
	echo_mode = ECHO_MASKED;
	break;

  case ECHO_MASKED:
	echo_mode = ECHO_DMASKED;
	break;

  case ECHO_DMASKED:
	echo_mode = ECHO_OFF;
	break;

  default:
	echo_mode = DEFAULT_ECHO_MODE;
	break;
  }

  if (debug_flag)
	fprintf(stderr, "Changed to echo_mode \"%s\"\n", echo_mode_str(echo_mode));

  set_echo_str(reply_buf);

  if (echo_mode != ECHO_OFF) {
	refresh_widget_buffer(echo_w, echo_buf);

	insertion_point = strlen(reply_buf);
	XawTextSetInsertionPoint(echo_w, insertion_point);
  }
  else
	refresh_widget_buffer(echo_w, echo_buf);

  XawTextDisplay(echo_w);
}

/*
 * Move the insertion point to the start of the line (reply buffer).
 */
/*ARGSUSED*/
static void
BeginningOfLine(Widget w, XEvent *event, String *args, Cardinal *nargs)
{
  XawTextPosition start_pos;

  if (debug_flag)
	fprintf(stderr, "BeginningOfLine\n");

  if (echo_mode == ECHO_OFF)
    start_pos = insertion_point;
  else
    start_pos = XawTextGetInsertionPoint(echo_w);

  if (start_pos > 0) {
	insertion_point = 0;
	if (echo_mode != ECHO_OFF)
	  XawTextSetInsertionPoint(echo_w, 0);
  }
}

/*
 * Move the insertion point to the end of the line (reply buffer).
 */
/*ARGSUSED*/
static void
EndOfLine(Widget w, XEvent *event, String *args, Cardinal *nargs)
{
  size_t len;

  if (debug_flag)
	fprintf(stderr, "EndOfLine\n");

  len = strlen(reply_buf);
  insertion_point = len;

  if (echo_mode != ECHO_OFF)
	XawTextSetInsertionPoint(echo_w, len);
}

/*
 * Delete the character immediately to the left of the cursor.
 */
/*ARGSUSED*/
static void
EraseChar(Widget w, XEvent *event, String *args, Cardinal *nargs)
{
  size_t len, rcount;
  XawTextPosition start_pos;

  if (echo_mode == ECHO_OFF)
	start_pos = insertion_point;
  else
	start_pos = XawTextGetInsertionPoint(echo_w);

  len = strlen(reply_buf);
  if (debug_flag)
	fprintf(stderr, "EraseChar @%ld/%ld\n", start_pos, len);

  if (len == 0)
	return;

  if (start_pos == 0 || start_pos > len)
	return;

  /*
   * Shift everything to the right of the insertion point one spot left,
   * unless the last character is being erased.
   */
  if ((rcount = len - start_pos) != 0) {
	memmove(echo_buf + start_pos - 1, echo_buf + start_pos, rcount);
	memmove(reply_buf + start_pos - 1, reply_buf + start_pos, rcount);
  }
  echo_buf[len - 1] = '\0';
  reply_buf[len - 1] = '\0';

  refresh_widget_buffer(w, echo_buf);

  insertion_point = start_pos - 1;
  if (echo_mode != ECHO_OFF)
	XawTextSetInsertionPoint(echo_w, start_pos - 1);

  XawTextDisplay(echo_w);
}

/*
 * Delete ("rubout") the character under the cursor.
 */
/*ARGSUSED*/
static void
DeleteChar(Widget w, XEvent *event, String *args, Cardinal *nargs)
{
  size_t len, rcount;
  XawTextPosition start_pos;

  if (echo_mode == ECHO_OFF)
	start_pos = insertion_point;
  else
	start_pos = XawTextGetInsertionPoint(echo_w);

  len = strlen(reply_buf);
  if (debug_flag)
	fprintf(stderr, "DeleteChar @%ld/%ld\n", start_pos, len);

  if (len == 0)
	return;

  if (start_pos >= len)
	return;

  if ((rcount = len - start_pos) != 0) {
	memmove(echo_buf + start_pos, echo_buf + start_pos + 1, rcount);
	memmove(reply_buf + start_pos, reply_buf + start_pos + 1, rcount);
  }
  echo_buf[len - 1] = '\0';
  reply_buf[len - 1] = '\0';

  refresh_widget_buffer(w, echo_buf);

  insertion_point = start_pos;
  if (echo_mode != ECHO_OFF)
	XawTextSetInsertionPoint(echo_w, start_pos);

  XawTextDisplay(echo_w);
}

/*
 * Erase the "word" immediately to the left of the insertion point (cursor).
 * a) Erase zero or more consecutive spaces and tabs (whitespace);
 * b) Erase zero or more consecutive non-whitespace characters;
 * c) Leave the insertion point at this position; and then
 * d) Update the echo buffer.
 *
 * This is intended to be equivalent to the tty handler's WERASE function.
 * and is much different from the behaviour of the original implementation's
 * erase-word() action... I have no idea what I was thinking at the
 * time wrt its usefulness.
 */
/*ARGSUSED*/
static void
EraseWord(Widget w, XEvent *event, String *args, Cardinal *nargs)
{
  size_t len, ndel, nleft;
  long pos;
  XawTextPosition left_pos, right_pos;

  /*
   * Find the current cursor position.
   * Character positions in a Text widget begin at 0 and end at n, where n is
   * the number of characters in the Text source widget.
   */
  if (echo_mode == ECHO_OFF)
	right_pos = insertion_point;
  else
	right_pos = XawTextGetInsertionPoint(echo_w);

  len = strlen(reply_buf);
  if (debug_flag)
	fprintf(stderr, "EraseWord: right_pos=%ld/%ld\n", right_pos, len);

  /*
   * If the reply buffer is empty or the cursor is at the start, there is
   * nothing to do.
   */
  if (len == 0 || right_pos == 0)
	return;

  /*
   * Scanning left from the current cursor position, determine how many
   * characters will be erased.
   */

  pos = right_pos - 1;
  /* Skip whitespace. */
  while (pos >= 0 && (reply_buf[pos] == ' ' || reply_buf[pos] == '\t'))
	pos--;

  /* Skip non-whitespace. */
  while (pos >= 0 && (reply_buf[pos] != ' ' && reply_buf[pos] != '\t'))
	pos--;

  /* We're looking at the start of the buffer or a whitespace character. */
  left_pos = pos;

  /*
   * Shift left everything to the right of the insertion point.
   */
  ndel = right_pos - left_pos - 1;
  nleft = len - ndel;
  if (right_pos == len)
	memset(reply_buf + left_pos + 1, 0, ndel);
  else
	memmove(reply_buf + left_pos + 1, reply_buf + right_pos, nleft);

  len -= ndel;

  reply_buf[len] = '\0';

  if (debug_flag)
	fprintf(stderr, "New reply buffer: \"%s\"\n", reply_buf);

  /* Change the widget's copy of the string. */
  set_echo_str(reply_buf);

  if (echo_mode != ECHO_OFF) {
	refresh_widget_buffer(echo_w, echo_buf);

	insertion_point = left_pos + 1;
	XawTextSetInsertionPoint(echo_w, insertion_point);
  }
  else
	refresh_widget_buffer(echo_w, echo_buf);

  XawTextDisplay(echo_w);
}

/*
 * Erase the entire contents of the reply and echo buffers for the current
 * prompt.
 */
/*ARGSUSED*/
static void
EraseLine(Widget w, XEvent *event, String *args, Cardinal *nargs)
{
  size_t len;

  len = strlen(echo_buf);
  if (debug_flag)
	fprintf(stderr, "EraseLine: %ld chars\n", len);

  /*
   * Ensure any sensitive information gets zapped.
   * Copies could possibly exist at lower levels...
   */
  memset(echo_buf, 0, app_resources.rlen + 1);
  memset(reply_buf, 0, app_resources.rlen + 1);

  insertion_point = 0;
  if (echo_mode != ECHO_OFF) {
	refresh_widget_buffer(echo_w, echo_buf);

	XawTextSetInsertionPoint(echo_w, 0);
	XawTextDisplay(echo_w);
  }

}

/*
 *
 */
/*ARGSUSED*/
static void
KillToEndOfLine(Widget w, XEvent *event, String *args, Cardinal *nargs)
{
  size_t len, ndel;
  XawTextPosition start_pos;

  len = strlen(echo_buf);
  if (debug_flag)
	fprintf(stderr, "KillToEndOfLine: %ld chars\n", len);

  /*
   * Find the current cursor position.
   * Character positions in a Text widget begin at 0 and end at n, where n is
   * the number of characters in the Text source widget.
   */
  if (echo_mode == ECHO_OFF)
	start_pos = insertion_point;
  else
	start_pos = XawTextGetInsertionPoint(echo_w);

  len = strlen(reply_buf);
  if (debug_flag)
	fprintf(stderr, "KillToEndOfLine: start_pos=%ld, len=%ld\n", start_pos, len);

  /*
   * If the reply buffer is empty or the cursor is at the end, there is
   * nothing to do.
   */
  if (len == 0)
	return;

  ndel = len - start_pos;
  if (ndel == 0)
	return;

  /*
   * Call XSetSelectionOwner() before setting CUT_BUFFER0 so that an existing
   * selection is overridden.
   */
  XSetSelectionOwner(XtDisplay(w), XA_PRIMARY, None, CurrentTime);

  /* Uses CUT_BUFFER0. */
  XStoreBytes(XtDisplay(w), reply_buf + start_pos, ndel);
  if (debug_flag)
	fprintf(stderr, "Paste: \"%s\"\n", reply_buf + start_pos);

  memset(reply_buf + start_pos + 1, 0, ndel);

  len -= ndel;

  reply_buf[len] = '\0';

  if (debug_flag)
	fprintf(stderr, "New reply buffer: \"%s\"\n", reply_buf);

  /* Change the widget's copy of the string. */
  set_echo_str(reply_buf);

  if (echo_mode != ECHO_OFF) {
	refresh_widget_buffer(echo_w, echo_buf);

	insertion_point = start_pos;
	XawTextSetInsertionPoint(echo_w, insertion_point);
  }
  else {
	refresh_widget_buffer(echo_w, echo_buf);
  }

  XawTextDisplay(echo_w);
}

static void
mask_echo_char(XawTextPosition pos)
{

  echo_buf[pos] = (echo_mode == ECHO_MASKED) ? echo_mask_char : echo_dmask_char;

  XawTextDisplay(echo_w);
}

/*
 * Callback function for XtAppAddTimeOut().
 * A timeout is automatically removed when it is triggered but it may also
 * be removed before being triggered by calling XtRemoveTimeOut().
 */
static void
dmask_callback(XtPointer data, XtIntervalId *timer)
{
  XawTextPosition pos;

  if (echo_dmask_pos != -1) {
	pos = (XawTextPosition) echo_dmask_pos;
	echo_dmask_pos = -1;

	if (debug_flag)
	  fprintf(stderr, "dmask_callback: timeout @%ld\n", pos);
	mask_echo_char(pos);
  }
  else {
	/* The timer has been cancelled and should be ignored. */
  }
}

/*
 * If there is room for one character (CH) in the reply buffer, insert it
 * at the reply buffer's insertion point, echo as appropriate, and return zero.
 * Note that the reply buffer will always be null-terminated.
 * Return -1 if the reply buffer is full.
 *
 * Redisplay of the echo buffer is not forced, so the caller may want to
 * do that.
 */
static int
insert_char(int ch)
{
  size_t len, rcount;
  XawTextPosition start_pos;

  len = strlen(reply_buf);
  if (len == app_resources.rlen) {
	/* The buffer is full. */
	if (debug_flag)
	  fprintf(stderr, "insert_char: reply buffer is full\n");
	return(-1);
  }

  if (echo_mode == ECHO_OFF) {
	start_pos = insertion_point;
	if (debug_flag)
	  fprintf(stderr, "insert_char: start_pos=%ld\n", start_pos);

	/* If this is not a simple append, shift the buffer one spot to the right. */
	if ((rcount = len - start_pos) != 0)
	  memmove(reply_buf + start_pos + 1, reply_buf + start_pos, rcount);
	reply_buf[start_pos] = ch;
	if (start_pos == len)
	  reply_buf[start_pos + 1] = '\0';

	if (debug_flag)
	  fprintf(stderr, "reply_buf: \"%s\"\n", reply_buf);

	insertion_point = start_pos + 1;
	return(0);
  }

  start_pos = XawTextGetInsertionPoint(echo_w);
  start_pos = insertion_point;
  if (debug_flag)
	fprintf(stderr, "insert_char start_pos=%ld\n", start_pos);

  /* If this is not a simple append, shift the buffer one spot to the right. */
  if ((rcount = len - start_pos) != 0) {
	memmove(echo_buf + start_pos + 1, echo_buf + start_pos, rcount);
	memmove(reply_buf + start_pos + 1, reply_buf + start_pos, rcount);
  }

  /* Append or insert the new character. */
  if (echo_mode == ECHO_ON)
	echo_buf[start_pos] = ch;
  else if (echo_mode == ECHO_MASKED)
	echo_buf[start_pos] = echo_mask_char;
  else if (echo_mode == ECHO_DMASKED) {
	/*
	 * If this character is typed before the previous one (if any) is
	 * masked, then the previous typed character must be masked now, the new
	 * character echoed, and the pending timeout cleared or replaced by one
	 * for the new character.
	 *
	 * The documentation doesn't say if there is an invalid XtIntervalId,
	 * so we cannot rely on the value of echo_dmask_timer to indicate
	 * whether a timeout is in progress.
	 * We use a non-negative echo_dmask_pos to indicate that a timeout
	 * is in progress for the character at that buffer position.
	 */
	if (echo_dmask_pos != -1) {
	  XawTextPosition pos;

	  XtRemoveTimeOut(echo_dmask_timer);
	  pos = (XawTextPosition) echo_dmask_pos;
	  echo_dmask_pos = -1;
	  mask_echo_char(pos);
	}
	echo_buf[start_pos] = ch;
	echo_dmask_pos = (long) start_pos;
	echo_dmask_timer = XtAppAddTimeOut(app_context,
								   (unsigned long) app_resources.dmask_duration,
									   dmask_callback, &echo_dmask_pos);
  }
  else {
	/* No echo. */
	echo_buf[start_pos] = '\0';
  }

  reply_buf[start_pos] = ch;
  if (start_pos == len)
	reply_buf[start_pos + 1] = '\0';
  if (debug_flag)
	fprintf(stderr, "reply_buf: \"%s\"\n", reply_buf);

  insertion_point++;
  XawTextSetInsertionPoint(echo_w, start_pos + 1);

  return(0);
}

/*
 * Insert a printable character at the current cursor position.
 * After error checking, this involves either appending the new character
 * to the buffer, or shifting the buffer contents to the right and stuffing
 * the new character at the current cursor position.
 *
 * Character positions in a Text widget begin at 0 and end at n, where n is
 * the number of characters in the Text source widget. 
 */
/*ARGSUSED*/
static void
InsertChar(Widget w, XEvent *event, String *args, Cardinal *nargs)
{
  int n;
  char buf[8];
  KeySym keysym;
  XKeyEvent *kev;

  kev = (XKeyEvent *) event;
  n = XLookupString(kev, buf, sizeof(buf), &keysym, NULL);
  if (n != 1 || !isprint((int) buf[0])) {
	/*
	 * This is not something to be inserted; e.g., XK_Shift_? or XK_Control_?
	 * so ignore it.
	 *
	 * Cursor motion and (most) control ops should be handled elsewhere by their
	 * default bindings.
	 * E.g., XK_Left, XK_Right, XK_Delete
	 */
	if (debug_flag)
	  fprintf(stderr, "InsertChar: ignoring keysym %lx\n", keysym);
	return;
  }

  if (insert_char((int) buf[0]) == -1)
	return;

  if (echo_mode == ECHO_OFF)
	return;

  if (debug_flag) {
	fprintf(stderr, "echo_buf: \"%s\"\n", echo_buf);
	fprintf(stderr, "reply_buf: \"%s\"\n", reply_buf);
   }

  refresh_widget_buffer(echo_w, echo_buf);

  if (echo_mode != ECHO_OFF)
	XawTextSetInsertionPoint(echo_w, insertion_point);

  XawTextDisplay(echo_w);
}

/*
 * Copy the entire reply buffer to CUT_BUFFER0, regardless of the echo mode.
 */
/*ARGSUSED*/
static void
SelectReply(Widget w, XEvent *event, String *args, Cardinal *nargs)
{

  if (debug_flag)
	fprintf(stderr, "SelectReply: \"%s\"\n", reply_buf);

  /*
   * Call XSetSelectionOwner() before setting CUT_BUFFER0 so that an existing
   * selection is overridden.
   */
  XSetSelectionOwner(XtDisplay(w), XA_PRIMARY, None, CurrentTime);

  /* Uses CUT_BUFFER0. */
  XStoreBytes(XtDisplay(w), reply_buf, (int) strlen(reply_buf));
  XFlush(XtDisplay(w));
}

/*
 * Fill as much as possible of reply_buf with printable text that has
 * been pasted from CUT_BUFFER0.  Any excess is silently discarded.
 */
/*ARGSUSED*/
static void
PasteSelection(Widget w, XEvent *event, String *args, Cardinal *nargs)
{
  int buffer, nbytes;
  char *line;

  if (debug_flag)
	fprintf(stderr, "PasteSelection\n");

  /*
   * Although there are eight cut buffers, we will keep this simple for now
   * and use only CUT_BUFFER0.
   * The buffer can contain arbitrary binary bytes, including nulls, so watch
   * for non-printable bytes.  [Some documentation says that they can only
   * contain text in the STRING encoding.]
   *
   * There is no support for Selections because, at first glance, it seems to
   * be an unnecessary complication.
   * See: Xlib Programming Manual, Volume One, Chapter 12, Interclient
   * Communication
   *
   * Buffer size limitations and null termination are enforced by insert_char().
   */
  buffer = 0;
  line = XFetchBuffer(XtDisplay(w), &nbytes, buffer);
  if (nbytes != 0 && line != NULL) {
	int i;
	char *p;

	fprintf(stderr, "Use CUT_BUFFER0: %d bytes:\n", nbytes);
	fprintf(stderr, "\"%s\"\n", line);
	for (p = line, i = 0; i < nbytes && isprint((int) *p); i++, p++) {
	  if (insert_char(*p) == -1)
		break;
	}
	/* XXX zap the memory? */
	XFree(line);
  }

  refresh_widget_buffer(echo_w, echo_buf);

  if (echo_mode != ECHO_OFF)
	XawTextSetInsertionPoint(echo_w, insertion_point);

  XawTextDisplay(echo_w);
}

/*
 * Normal termination.
 * Print the response(s) to stdout, undo any changes made at startup,
 * then exit.
 */
/*ARGSUSED*/
static void
FinishPrompt(Widget w, XEvent *event, String *args, Cardinal *nargs)
{
  int i;

  if (debug_flag)
	fprintf(stderr, "FinishPrompt\n");

  strcpy(npa->reply_buf, reply_buf);

  for (i = 0, npa = prompt_args; i < nprompts; npa = npa->next, i++)
	printf("%s\n", npa->reply_buf);

  unwarp();
  XtDestroyWidget(toplevel_w);

  exit(0);
}

/*ARGSUSED*/
static void
PreviousPrompt(Widget w, XEvent *event, String *args, Cardinal *nargs)
{

  if (debug_flag)
	fprintf(stderr, "PreviousPrompt\n");

  npa_prev = npa;
  npa = npa->prev;
  if (--current_prompt == 0)
	current_prompt = nprompts;

  exchange_prompt();
}

/*ARGSUSED*/
static void
NextPrompt(Widget w, XEvent *event, String *args, Cardinal *nargs)
{

  if (debug_flag)
	fprintf(stderr, "NextPrompt\n");

  if (nprompts > 1) {
	npa_prev = npa;
	npa = npa->next;
	if (++current_prompt > nprompts)
	  current_prompt = 1;

	exchange_prompt();
  }
}

/*ARGSUSED*/
static void
NextPromptOrFinish(Widget w, XEvent *event, String *args, Cardinal *nargs)
{

  if (debug_flag)
	fprintf(stderr, "NextPromptOrFinish\n");

  if (nprompts_seen == nprompts && app_resources.return_exit) {
	FinishPrompt(w, event, args, nargs);
	/*NOTREACHED*/
  }

  NextPrompt(w, event, args, nargs);
}

/*ARGSUSED*/
static void
Abort(Widget w, XEvent *event, String *args, Cardinal *nargs)
{

  if (debug_flag)
	fprintf(stderr, "Abort\n");

  unwarp();
  XtDestroyWidget(toplevel_w);

  exit(1);
}

/*ARGSUSED*/
static void
Ignore(Widget w, XEvent *event, String *args, Cardinal *nargs)
{

  if (debug_flag)
	fprintf(stderr, "Ignore\n");
}

/*ARGSUSED*/
static void
VisibilityEvent(Widget w, XEvent *event, String *args, Cardinal *nargs)
{

  if (debug_flag)
	fprintf(stderr, "VisibilityEvent\n");

  visible = 1;
}

/* Map actions to handler functions. */
static XtActionsRec xprompt_actions[] = {
  /* Prompt change & insertion point movement. */
  {"xp-backward-char",         BackwardChar },
  {"xp-beginning-of-line",     BeginningOfLine },
  {"xp-end-of-line",           EndOfLine },
  {"xp-finish-prompt",		   FinishPrompt },
  {"xp-forward-char",          ForwardChar },
  {"xp-next-prompt",		   NextPrompt },
  {"xp-next-prompt-or-finish", NextPromptOrFinish },
  {"xp-previous-prompt",	   PreviousPrompt },

  {"xp-select-reply",		   SelectReply },

  /* Insert & delete. */
  {"xp-delete-char",           DeleteChar },
  {"xp-erase-char",            EraseChar },
  {"xp-erase-line",		       EraseLine },
  {"xp-erase-word",            EraseWord },
  {"xp-insert-char",           InsertChar },
  {"xp-paste-selection",       PasteSelection },
  {"xp-kill-to-end-of-line",   KillToEndOfLine },

  /* Misc. actions. */
  {"xp-abort",				   Abort },
  {"xp-toggle-echo-mode",      ToggleEchoMode },
  {"xp-ignore",                Ignore },
  {"xp-redraw-display",        RedrawDisplay },
  {"visibility-event",         VisibilityEvent },
};

static void
usage(int version_only_flag)
{

  fprintf(stderr, "xprompt %s [%s]\n",
		  XPROMPT_VERSION_RELEASE, XPROMPT_VERSION_DATE);
  if (version_only_flag)
	exit(0);

  fprintf(stderr, "%s\n", revid);

  fprintf(stderr, "Usage:\n");
  fprintf(stderr, "%s [flags] [xtoolkitargs] -p prompt [-r reply] \
[-p prompt [-r reply]] ...\n",
		  progname);
  fprintf(stderr, "where <flags> is one or more of:\n");
  fprintf(stderr, "-h           Print help and exit [also --help]\n");
  fprintf(stderr, "-v           Print version and exit [also --version]\n");
  fprintf(stderr, "-rlen #      Maximum length of any reply (default %d)\n",
		  DEFAULT_MAXIMUM_REPLY_LEN);
  fprintf(stderr, "-[no]debug   Emit debugging (%s)\n",
		  DEFAULT_DEBUG ? "debug" : "nodebug");
  fprintf(stderr,
		  "-echo <mode> How to echo reply text (%s) [on|off|mask|dmask]\n",
		  echo_mode_str(DEFAULT_ECHO_MODE));
  fprintf(stderr, "-dmd <ms>    For dmask echo mode, use a <ms> delay (%d)\n",
		  DEFAULT_DMASK_DURATION);
  fprintf(stderr, "-[no]grab    Grab keyboard (%s)\n",
		  DEFAULT_GRAB ? "grab" : "nograb");
  fprintf(stderr,
		  "-[no]re      Return/Enter key exits if there's one prompt (%s)\n",
		  DEFAULT_RETURN_EXIT ? "re" : "nore");
  fprintf(stderr, "-[no]warp    Warp cursor to reply window (%s)\n",
		  DEFAULT_WARP ? "warp" : "nowarp");
  fprintf(stderr, "-pfn <pfont> Use <pfont> for prompts\n");
  fprintf(stderr, "-efn <rfont> Use <rfont> for echoing replies\n");
  fprintf(stderr, "-ibw #       Border width of inside window (%d)\n",
		  DEFAULT_INSIDE_BORDER_WIDTH);
  fprintf(stderr, "-tf <file>   Translation file to override defaults\n");
  fprintf(stderr, "-p <prompt>  Add <prompt> to the list of prompts\n");
  fprintf(stderr, "-r <reply>   Use <reply> for the prompt's default reply\n");

  exit(1);
}

int
main(int argc, char **argv)
{
  int a, i, len, max_user_prompt_len;
  int geom_mask, geom_x, geom_y;
  int prompt_bb_width, prompt_bb_height, echo_bb_width, echo_bb_height;
  unsigned int geom_height, geom_width;
  char geom_str[100];
  Arg arg[20];
  Widget box, popup;
  XtTranslations t;
  XFontStruct *prompt_font, *echo_font;

  progname = argv[0];
  echo_dmask_pos = -1;
  insertion_point = 0;

  /* Do this as early as possible, just in case... but also later. */
  if (argv[1] == NULL
	  || strcaseeq(argv[1], "-h") || strcaseeq(argv[1], "--help")) {
	usage(0);
	/*NOTREACHED*/
  }
  if (strcaseeq(argv[1], "-v") || strcaseeq(argv[1], "--version")) {
	usage(1);
	/*NOTREACHED*/
  }

  /*
   * applicationShellWidgetClass (toplevel_w, "XPrompt"):
   *   topLevelShellWidgetClass (popup, "xprompt"):
   *     boxWidgetClass (box, "box")):
   *       asciiTextWidgetClass (prompt_w, "text"):
   *       asciiTextWidgetClass (echo_w, "echo"):
   */

  toplevel_w = XtOpenApplication(&app_context, "XPrompt",
							   option_table, XtNumber(option_table),
							   &argc, argv, NULL, applicationShellWidgetClass,
							   NULL, 0);

  XtGetApplicationResources(toplevel_w, &app_resources, resources,
							XtNumber(resources), NULL, (Cardinal) 0);

  /* Disable core dumps, which could conceivably contain a password... */
  disable_dump();

  if (app_resources.rlen < 1) {
	fprintf(stderr, "xprompt: -rlen value must be positive\n");
	usage(0);
	/*NOTREACHED*/
  }

  if (app_resources.insideborderwidth < 0) {
	fprintf(stderr, "xprompt: -ibw value must not be negative\n");
	usage(0);
	/*NOTREACHED*/
  }

  if (app_resources.dmask_duration < 0) {
	fprintf(stderr, "xprompt: -dmd value must not be negative\n");
	usage(0);
	/*NOTREACHED*/
  }

  /*
   * Handle any command line flags that were not specified in option_table
   * and processed by XtOpenApplication().
   */
  nprompts = 0;
  prompt_args = NULL;
  max_user_prompt_len = 0;
  for (i = 1; i < argc; i++) {
	if (strcaseeq(argv[i], "-v") || strcaseeq(argv[i], "--version")) {
	  usage(1);
	  /*NOTREACHED*/
	}
	if (strcaseeq(argv[i], "-h") || strcaseeq(argv[i], "--help")) {
	  usage(0);
	  /*NOTREACHED*/
	}

	/* Expect to see a prompt, optionally followed by a default reply. */
	if (i == argc - 1 || !streq(argv[i], "-p"))
	  usage(0);

	if (nprompts == MAX_PROMPTS) {
	  fprintf(stderr, "xprompt: too many prompts\n");
	  usage(0);
	  /*NOTREACHED*/
	}

	if (prompt_args) {
	  npa->next = (Prompt_args *) malloc(sizeof(Prompt_args));
	  npa->next->prev = npa;
	  npa = npa->next;
	  npa->next = NULL;
	}
	else {
	  prompt_args = npa = (Prompt_args *) malloc(sizeof(Prompt_args));
	  npa->prev = npa->next = NULL;
	}

	npa->prompt_buf = argv[++i];
	if ((len = strlen(npa->prompt_buf)) > max_user_prompt_len)
	  max_user_prompt_len = len;

	if ((npa->reply_buf = (char *) malloc(app_resources.rlen + 1)) == NULL) {
	  fprintf(stderr, "Can't malloc reply buffer\n");
	  exit(1);
	}
	npa->reply_buf[0] = '\0';

	if (argv[i + 1] && streq(argv[i + 1], "-r")) {
	  if (++i == argc - 1) {
		usage(0);
		/*NOTREACHED*/
	  }

	  if (strlen(argv[++i]) > app_resources.rlen) {
		fprintf(stderr, "xprompt: default reply is too long\n");
		exit(1);
	  }
	  strcpy(npa->reply_buf, argv[i]);
	}

	npa->seenprompt = FALSE;
	nprompts++;
  }

  /* For convenience, some global flags. */
  debug_flag = app_resources.debug;
  echo_mode = echo_mode_arg(app_resources.echo_mode_str);
  if (echo_mode == ECHO_INVALID) {
	fprintf(stderr, "Unrecognized echo mode.\n");
	usage(0);
	/*NOTREACHED*/
  }

  echo_mask_char = DEFAULT_ECHO_MASK_CHAR;
  if (app_resources.echo_mask_char != NULL)
	echo_mask_char = *app_resources.echo_mask_char;
  echo_dmask_char = DEFAULT_ECHO_DMASK_CHAR;
  if (app_resources.echo_dmask_char != NULL)
	echo_dmask_char = *app_resources.echo_dmask_char;

  if (nprompts == 0) {
	/* No prompts? */
	usage(0);
	/*NOTREACHED*/
  }

  /*
   * If there is more than one prompt, each prompt string includes an
   * identifying suffix.  Account for that by allocating a sufficient
   * amount extra.
   * See exchange_prompt().
   */
  max_displayed_prompt_len
	= displayed_prompt_length(nprompts, max_user_prompt_len);
  prompt_buf = (char *) malloc(max_displayed_prompt_len + 1);

  if (debug_flag) {
	char *tt;

	fprintf(stderr, "XawVersion: %ld\n", XawVersion); 

	fprintf(stderr, "nprompts=%d\n", nprompts);
	fprintf(stderr, "echo_mode=\"%s\"\n", echo_mode_str(echo_mode));
	fprintf(stderr, "echo_mask_char=\"%c\"\n", echo_mask_char);
	fprintf(stderr, "echo_dmask_char=\"%c\"\n", echo_dmask_char);
	fprintf(stderr, "dmask_duration=%dms\n", app_resources.dmask_duration);
	fprintf(stderr, "rlen=%d\n", app_resources.rlen);
	fprintf(stderr, "grab=%d\n", app_resources.grab);
	fprintf(stderr, "warp=%d\n", app_resources.warp);
	fprintf(stderr, "return_exit=%d\n", app_resources.return_exit);
	fprintf(stderr, "geometry=\"%s\"\n", app_resources.geometry);
	fprintf(stderr, "borderwidth=%d\n", app_resources.borderwidth);
	fprintf(stderr, "insideborderwidth=%d\n", app_resources.insideborderwidth);
	fprintf(stderr, "max_user_prompt_len=%d\n", max_user_prompt_len);
	fprintf(stderr, "max_displayed_prompt_len=%d\n", max_displayed_prompt_len);

	fprintf(stderr, "Text translations:\n");
	tt = strdup(xprompt_TextTranslations);
	while (*tt != '\0') {
	  char *p;

	  if ((p = strchr(tt, (int) '\n')) == NULL)
		break;
	  *p++ = '\0';
	  fprintf(stderr, "%s\n", tt);
	  tt = p;
	}
  }

  prompt_args->prev = npa;
  npa->next = prompt_args;
  npa = prompt_args;
  current_prompt = 1;

  reply_buf = (char *) malloc(app_resources.rlen + 1);
  memset(reply_buf, 0, app_resources.rlen + 1);

  echo_buf = (char *) malloc(app_resources.rlen + 1);
  memset(echo_buf, 0, app_resources.rlen + 1);

  if (app_resources.pfont == NULL)
	app_resources.pfont = app_resources.font;
  if (app_resources.efont == NULL)
	app_resources.efont = app_resources.font;

  XtAppAddActions(app_context, xprompt_actions, XtNumber(xprompt_actions));

  /*
   * Determine the window geometry.
   */
  prompt_font = app_resources.pfont;
  prompt_bb_width = prompt_font->max_bounds.rbearing
	- prompt_font->min_bounds.lbearing;
  prompt_bb_height = prompt_font->max_bounds.ascent
	+ prompt_font->max_bounds.descent;
  echo_font = app_resources.efont;
  echo_bb_width = echo_font->max_bounds.rbearing
	- echo_font->min_bounds.lbearing;
  echo_bb_height = echo_font->max_bounds.ascent
	+ echo_font->max_bounds.descent;

  geom_x = DEFAULT_GEOM_X;				/* An arbitrary default location. */
  geom_y = DEFAULT_GEOM_Y;
  geom_width = max_displayed_prompt_len * prompt_bb_width
	+ app_resources.rlen * echo_bb_width + GEOM_WIDTH_PAD;
  geom_height = MAXIMUM(prompt_bb_height, echo_bb_height) + GEOM_HEIGHT_PAD;
  geom_mask = XParseGeometry(app_resources.geometry, &geom_x, &geom_y,
							 &geom_width, &geom_height);
  if (geom_mask & WidthValue)
	geom_width *= echo_bb_width;
  if (geom_mask & HeightValue)
	geom_height *= echo_bb_height;
  make_geometry_str(geom_str, sizeof(geom_str),
					geom_mask, geom_width, geom_height, geom_x, geom_y);

  if (debug_flag) {
	fprintf(stderr, "geom_x=%d geom_y=%d\n", geom_x, geom_y);
	fprintf(stderr, "prompt_bb_width=%d prompt_bb_height=%d\n",
			prompt_bb_width, prompt_bb_height);
	fprintf(stderr, "echo_bb_width=%d echo_bb_height=%d\n",
			echo_bb_width, echo_bb_height);
	fprintf(stderr, "geom_width=%d geom_height=%d\n",
			max_displayed_prompt_len * prompt_bb_width
			+ app_resources.rlen * echo_bb_width + GEOM_WIDTH_PAD,
			MAXIMUM(prompt_bb_height, echo_bb_height) + GEOM_HEIGHT_PAD);
	fprintf(stderr, "geom_mask=%d geom_width=%d geom_height=%d\n",
			geom_mask, geom_width, geom_height);
	fprintf(stderr, "geom_str=\"%s\"\n", geom_str);
  }

  /*
   * Note:
   * XtSetArg is a macro that may evaluate its first argument more than once,
   * so beware side effects, such as autoincrementing a counter too many times...
   * We'll use this idiom to avoid this issue.
   */
  a = 0;
  XtSetArg(arg[a], XtNgeometry, geom_str); a++;
  XtSetArg(arg[a], XtNborderWidth, app_resources.borderwidth); a++;

  /*
   * If you prefer a popup window instead of a managed window, you
   * can use overrideShellWidgetClass instead of topLevelShellWidgetClass:
   *    popup = XtCreatePopupShell("popup", overrideShellWidgetClass,
   *			toplevel_w, arg, (Cardinal) 2);
   */
  popup = XtCreatePopupShell("xprompt", topLevelShellWidgetClass,
							 toplevel_w, arg, (Cardinal) a);

  box = XtCreateManagedWidget("box", boxWidgetClass,
							  popup, NULL, (Cardinal) 0);

  /* Initialize the first prompt. */
  set_prompt(prompt_args->prompt_buf, prompt_args->reply_buf, 1, nprompts);

  prompt_args->seenprompt = TRUE;
  nprompts_seen = 1;

  a = 0;
  XtSetArg(arg[a], XtNstring, prompt_buf); a++;
  XtSetArg(arg[a], XtNlength, strlen(prompt_buf)); a++;
  XtSetArg(arg[a], XtNborderWidth, 0); a++;
  XtSetArg(arg[a], XtNfont, app_resources.pfont); a++;
  XtSetArg(arg[a], XtNeditType, XawtextRead); a++;
  XtSetArg(arg[a], XtNwidth,
		   prompt_bb_width * (max_displayed_prompt_len + 1)); a++;
  XtSetArg(arg[a], XtNsensitive, False); a++;
  XtSetArg(arg[a], XtNuseStringInPlace, True); a++;
  XtSetArg(arg[a], XtNdisplayCaret, False); a++;
  XtSetArg(arg[a], XtNtype, XawAsciiString); a++;
  prompt_w = XtCreateManagedWidget("text", asciiTextWidgetClass,
								   box, arg, (Cardinal) a);

  /*
   * Initialize the default reply to the first prompt.
   */

  /*
   * Athena Widget documentation for X11R6.4 libXaw 1.0.7 says:
   * If useStringInPlace is True for an asciiTextWidgetClass widget, the
   * application is the "string owner", the widget takes the value of the
   * string resource as its own text buffer, and its length resource indicates
   * the buffer size.
   *
   * It seems that there might be an off-by-one bug in libXaw or a problem
   * with its documentation, both of which are difficult to believe.  But I
   * observed that when a buffer length of 6 was given as XtNlength, libXaw
   * pasted (via an insert-selection(PRIMARY) text translation) 7 characters of
   * text into the buffer, the last character of which was from the pasted text
   * rather than a null byte.
   * So this could (and did) leave the buffer unterminated, which can't be
   * correct.
   *
   * To avoid this here, we allocate an extra byte, making sure it is always
   * NUL, and tell libXaw that the buffer size is two bytes less than what has
   * actually been allocated.
   */

  a = 0;
  XtSetArg(arg[a], XtNscrollVertical, XawtextScrollWhenNeeded); a++;
  XtSetArg(arg[a], XtNheight, echo_bb_height + 6); a++;
  XtSetArg(arg[a], XtNstring, echo_buf); a++;
  XtSetArg(arg[a], XtNlength, app_resources.rlen - 1); a++;
  /*
   * Text can be modified only explicitly through XtSetValues, not directly by
   * the user (pasting/insert-selection is also disabled in this mode).
   * editType XawtextEdit would make the text fully editable.
   */
  XtSetArg(arg[a], XtNeditType, XawtextRead); a++;
  XtSetArg(arg[a], XtNborderWidth, app_resources.insideborderwidth); a++;
  XtSetArg(arg[a], XtNwidth, app_resources.rlen * echo_bb_width); a++;
  XtSetArg(arg[a], XtNinsertPosition, strlen(echo_buf)); a++;
  XtSetArg(arg[a], XtNfont, app_resources.efont); a++;
  XtSetArg(arg[a], XtNuseStringInPlace, True); a++;
  XtSetArg(arg[a], XtNtype, XawAsciiString); a++;
  /*
  XtSetArg(arg[a], XtNecho, echo_flag ? True : False); a++;
  */
  echo_w = XtCreateManagedWidget("echo", asciiTextWidgetClass,
								 box, arg, (Cardinal) a);

  /*
   * We need to echo input according to the Echo_mode in effect and
   * the behaviour of a particular action may change depending on ECHO_MODE
   * (and be different from the normal action).
   * We need to implement many of our own actions and catch all input.
   * Some standard text actions are unsupported and may cause incorrect
   * behaviour or crash the program.
   * A user is free to change bindings but should not revert a custom
   * action to standard X Toolkit action.
   *
   * Compile and install/replace the widget's translations.
   */
  t = XtParseTranslationTable(xprompt_TextTranslations);
  if (t == NULL) {
	fprintf(stderr, "xprompt: error parsing built-in translation table.\n");
	exit(1);
  }
  XtOverrideTranslations(echo_w, t);

  if (app_resources.texttranslations != NULL) {
	if (debug_flag)
	  fprintf(stderr, "Parsing resource text translations:\n\"%s\"\n",
			  app_resources.texttranslations);
	t = XtParseTranslationTable(app_resources.texttranslations);
	if (t == NULL) {
	  fprintf(stderr, "xprompt: error parsing resource text translations.\n");
	  exit(1);
	}
	XtOverrideTranslations(echo_w, t);
  }

  /* Incorporate any translations specified by the user from a file. */
  if (app_resources.texttranslationfile != NULL
	  && load_user_text_translations(app_resources.texttranslationfile) == -1)
	exit(1);

  visible = 0;
  XtPopup(popup, XtGrabNonexclusive);

  /*
   * Handle the grab and/or warp, if necessary.
   */
  if (app_resources.grab || app_resources.warp) {
	XWindowAttributes echo_w_attr;

	if (XGetWindowAttributes(XtDisplay(echo_w), XtWindow(echo_w),
							 &echo_w_attr) == 0)
	  app_resources.warp = app_resources.grab = FALSE;
	else if (!visible) {
	  XEvent event;
	  long em;

	  /*
	   * The following weirdness waits for the windows to become visible
	   * before warping the cursor or setting the input focus.
	   * This is necessary to avoid the cases where a window manager
	   * repositions things *after* the warp has occurred and where the
	   * focus window is not viewable at the time of the XSetInputFocus.
	   * Suggestions on the Right Thing welcome.
	   */
	  em = VisibilityChangeMask;
	  XSelectInput(XtDisplay(popup), XtWindow(popup), em);
	  while (1) {
		bzero(&event, sizeof(event));
		XNextEvent(XtDisplay(popup), &event);
		if (event.type == VisibilityNotify)
		  break;
	  }
	  em = echo_w_attr.all_event_masks;
	  XSelectInput(XtDisplay(popup), XtWindow(popup), em);
	}

	if (app_resources.grab) {
	  XGrabKeyboard(XtDisplay(echo_w), XtWindow(echo_w), False,
					GrabModeAsync, GrabModeAsync, CurrentTime);
	  XSetInputFocus(XtDisplay(echo_w), XtWindow(echo_w),
					 RevertToPointerRoot, CurrentTime);
	}

	if (app_resources.warp) {
	  Window root_return, child_return;
	  int root_x_return, root_y_return;
	  int win_x_return, win_y_return;
	  int reply_x, reply_y;
	  unsigned int mask_return;

	  warp_info.orig_x = -1;
	  warp_info.orig_y = -1;
	  /*
	   * Find out where the cursor is so that it can be put back
	   * before the program exits.
	   */
	  if (XQueryPointer(XtDisplay(popup), XtWindow(popup),
						&root_return, &child_return,
						&root_x_return, &root_y_return,
						&win_x_return, &win_y_return,
						&mask_return) != 0) {
		warp_info.orig_win = root_return;
		warp_info.orig_x = root_x_return;
		warp_info.orig_y = root_y_return;
	  }
	  reply_x = echo_w_attr.width - (echo_w_attr.width / 10);
	  reply_y = echo_w_attr.height / 2;

	  XWarpPointer(XtDisplay(echo_w), None, XtWindow(echo_w),
				   0, 0, 0, 0, reply_x, reply_y);
	}
  }

  XtAppMainLoop(app_context);
}
